from __future__ import unicode_literals

import six
from unittest import TestCase
from ..inmemory import ChannelLayer
from ..wsgi import WsgiToAsgiAdapter


class TestWsgiToAsgiAdapter(WsgiToAsgiAdapter):
    """
    Testing subclass with a fixed reply channel
    """

    def get_reply_channel(self):
        return "!http.response.testtest"


class WsgiToAsgiTests(TestCase):
    """
    Tests the WSGI-to-ASGI adapter.
    """

    def setUp(self):
        """
        Make an in memory channel layer for testing
        """
        self.channel_layer = ChannelLayer()
        self.reply_channel = "!http.response.testtest"
        self.start_response_value = None
        self.application = TestWsgiToAsgiAdapter(self.channel_layer)
    
    def native_string(self, value):
        """
        Makes sure that the passed in string value comes out as a PEP3333
        "native string".
        """
        if six.PY2:
            if isinstance(value, unicode):
                return value.encode("latin1")
            else:
                return value
        else:
            if isinstance(value, bytes):
                return value.decode("latin1")
            else:
                return value

    def start_response(self, status, headers, exc_info=None):
        self.start_response_value = [status, headers, exc_info]

    def test_basic(self):
        # Example request
        ns = self.native_string
        environ = {
            "PATH_INFO": ns("/"),
            "CONTENT_TYPE": ns("text/html; charset=utf-8"),
            "REQUEST_METHOD": ns("GET"),
        }
        # Inject the response ahead of time
        self.channel_layer.send(self.reply_channel, {
            "status": 200,
            "content": b"Hi there!",
        })
        # Run WSGI adapter
        response = list(self.application(environ, self.start_response))
        # Make sure response looks fine
        self.assertEqual(response[0], b"Hi there!")
        self.assertEqual(self.start_response_value[0], b"200 OK")



"""
ASGI spec conformance test suite.

Calling the functions with an ASGI channel layer instance will return you a
single TestCase instance that checks for conformity on that instance.

You MUST also pass along an expiry value to the sets of tests, to allow the
suite to wait correctly for expiry. It's suggested you configure the layer
for 1-second expiry during tests, and use a 1.1 second expiry delay.

The channel layers should be empty to start with, and discarded after use,
as they'll be full of test data. If a layer supports the "flush" extension,
it'll be flushed before every test.
"""

from __future__ import unicode_literals
import six
import time
import unittest


def make_tests(channel_layer, expiry_delay):

    class LayerTestCase(unittest.TestCase):
        """
        Tests that core ASGI functionality is maintained.
        """

        def setUp(self):
            if "flush" in channel_layer.extensions:
                channel_layer.flush()

        def test_send_recv(self):
            """
            Tests that channels can send and receive messages right.
            """
            channel_layer.send("sr_test", {"value": "blue"})
            channel_layer.send("sr_test", {"value": "green"})
            channel_layer.send("sr_test2", {"value": "red"})
            # Get just one first
            channel, message = channel_layer.receive_many(["sr_test"])
            self.assertEqual(channel, "sr_test")
            self.assertEqual(message, {"value": "blue"})
            # And the second
            channel, message = channel_layer.receive_many(["sr_test"])
            self.assertEqual(channel, "sr_test")
            self.assertEqual(message, {"value": "green"})
            # And the other channel with multi select
            channel, message = channel_layer.receive_many(["sr_test", "sr_test2"])
            self.assertEqual(channel, "sr_test2")
            self.assertEqual(message, {"value": "red"})

        def test_unicode_channel_name(self):
            """
            Makes sure channel names can handle unicode
            """
            channel_layer.send("\u00a3_test", {"value": "blue"})
            # Get just one first
            channel, message = channel_layer.receive_many(["\u00a3_test"])
            self.assertEqual(channel, "\u00a3_test")
            self.assertEqual(message, {"value": "blue"})

        def test_message_expiry(self):
            """
            Tests that messages expire correctly.
            """
            channel_layer.send("me_test", {"value": "blue"})
            time.sleep(expiry_delay)
            channel, message = channel_layer.receive_many(["me_test"])
            self.assertIs(channel, None)
            self.assertIs(message, None)

        def test_new_channel(self):
            """
            Tests that new channel names are made correctly.
            """
            pattern = "test.foo!"
            name1 = channel_layer.new_channel(pattern)
            self.assertFalse(name1.endswith("!"))
            self.assertTrue("!" in name1)
            self.assertEqual(name1.find("!"), name1.rfind("!"))
            self.assertIsInstance(name1, six.text_type)
            # Send a message and make sure new_channel on second pass changes
            channel_layer.send(name1, {"value": "blue"})
            name2 = channel_layer.new_channel(pattern)
            # Make sure we can consume off of that new channel
            channel, message = channel_layer.receive_many([name1, name2])
            self.assertEqual(channel, name1)
            self.assertEqual(message, {"value": "blue"})
            # Make sure that new_channel rejects bad patterns
            with self.assertRaises(Exception):
                channel_layer.new_channel("test!foo")
            with self.assertRaises(Exception):
                channel_layer.new_channel("test.foo")

        def test_strings(self):
            """
            Ensures byte strings and unicode strings both make it through
            serialization properly.
            """
            # Message. Double-nested to ensure serializers are recursing properly.
            message = {
                "values": {
                    # UTF-8 sequence for british pound, but we want it not interpreted into that.
                    "utf-bytes": b"\xc2\xa3",
                    # Actual unicode for british pound, should come back as 1 char
                    "unicode": "\u00a3",
                    # Emoji, in case someone is using 3-byte-wide unicode storage
                    "emoji": "\u1F612",
                    # Random control characters and null
                    "control": b"\x01\x00\x03\x21",
                }
            }
            # Send it and receive it
            channel_layer.send("str_test", message)
            _, received = channel_layer.receive_many(["str_test"])
            # Compare
            self.assertIsInstance(received["values"]["utf-bytes"], six.binary_type)
            self.assertIsInstance(received["values"]["unicode"], six.text_type)
            self.assertIsInstance(received["values"]["emoji"], six.text_type)
            self.assertIsInstance(received["values"]["control"], six.binary_type)
            self.assertEqual(received["values"]["utf-bytes"], message["values"]["utf-bytes"])
            self.assertEqual(received["values"]["unicode"], message["values"]["unicode"])
            self.assertEqual(received["values"]["emoji"], message["values"]["emoji"])
            self.assertEqual(received["values"]["control"], message["values"]["control"])

        @unittest.skipIf("groups" not in channel_layer.extensions, "No groups extension")
        def test_groups(self):
            """
            Tests that basic group addition and send works
            """
            # Make a group and send to it
            channel_layer.group_add("tgroup", "tg_test")
            channel_layer.group_add("tgroup", "tg_test2")
            channel_layer.group_add("tgroup", "tg_test3")
            channel_layer.group_discard("tgroup", "tg_test3")
            channel_layer.send_group("tgroup", {"value": "orange"})
            # Receive from the two channels in the group and ensure messages
            channel, message = channel_layer.receive_many(["tg_test"])
            self.assertEqual(channel, "tg_test")
            self.assertEqual(message, {"value": "orange"})
            channel, message = channel_layer.receive_many(["tg_test2"])
            self.assertEqual(channel, "tg_test2")
            self.assertEqual(message, {"value": "orange"})
            # Make sure another channel does not get a message
            channel, message = channel_layer.receive_many(["tg_test3"])
            self.assertIs(channel, None)
            self.assertIs(message, None)

        @unittest.skipIf("flush" not in channel_layer.extensions, "No flush extension")
        def test_flush(self):
            """
            Tests that messages go away after a flush.
            """
            channel_layer.send("fl_test", {"value": "blue"})
            channel_layer.flush()
            channel, message = channel_layer.receive_many(["fl_test"])
            self.assertIs(channel, None)
            self.assertIs(message, None)

        @unittest.skipIf("flush" not in channel_layer.extensions, "No flush extension")
        @unittest.skipIf("groups" not in channel_layer.extensions, "No groups extension")
        def test_flush_groups(self):
            """
            Tests that groups go away after a flush.
            """
            channel_layer.send("fl_test", {"value": "blue"})
            channel_layer.flush()
            channel, message = channel_layer.receive_many(["fl_test"])
            self.assertIs(channel, None)
            self.assertIs(message, None)

    return LayerTestCase



import os
import sys
from setuptools import find_packages, setup
from asgiref import __version__


# We use the README as the long_description
readme_path = os.path.join(os.path.dirname(__file__), "README.rst")


setup(
    name='asgiref',
    version=__version__,
    url='http://github.com/andrewgodwin/asgiref/',
    author='Andrew Godwin',
    author_email='andrew@aeracode.org',
    description='Reference ASGI adapters and channel layers',
    long_description=open(readme_path).read(),
    license='BSD',
    zip_safe=False,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'six',
    ]
)



from __future__ import unicode_literals

import random
import six
import string
import time
import threading
from collections import deque


# TODO: Ensure thread safety

class ChannelLayer(object):
    """
    In memory channel layer object; a single one is instantiated as
    "channel_layer" for easy shared use.
    """

    def __init__(self, expiry=60):
        self.expiry = expiry
        self.thread_lock = threading.Lock()
        # Storage for state
        self._channels = {}
        self._groups = {}

    ### ASGI API ###

    extensions = ["groups", "flush"]

    class MessageTooLarge(Exception):
        pass

    def send(self, channel, message):
        # Make sure the message is a dict at least (no deep inspection)
        assert isinstance(message, dict), "message is not a dict"
        # Channel name should be text
        assert isinstance(channel, six.text_type), "%s is not text" % channel
        # Add it to a deque for the appropriate channel name
        with self.thread_lock:
            self._channels.setdefault(channel, deque()).append((
                time.time() + self.expiry,
                message,
            ))

    def receive_many(self, channels, block=False):
        # Shuffle channel names to ensure approximate global ordering
        channels = list(channels)
        random.shuffle(channels)
        # Expire old messages
        self._clean_expired()
        # Go through channels and see if a message is available:
        with self.thread_lock:
            for channel in channels:
                if self._channels.get(channel, None):
                    _, message = self._channels[channel].popleft()
                    return channel, message
        # No message available
        return None, None

    def new_channel(self, pattern):
        assert isinstance(pattern, six.text_type)
        assert pattern.endswith("!"), "New channel pattern must end with !"
        # Keep making channel names till one isn't present.
        while True:
            random_string = "".join(random.choice(string.ascii_letters) for i in range(8))
            new_name = pattern + random_string
            # Basic check for existence
            if new_name not in self._channels:
                return new_name

    ### ASGI Group API ###

    def group_add(self, group, channel):
        # Both should be text
        assert isinstance(group, six.text_type), "%s is not text" % group
        assert isinstance(channel, six.text_type), "%s is not text" % channel
        # Add to group set
        self._groups[group] = self._groups.get(group, set()).union({channel})

    def group_discard(self, group, channel):
        # Both should be text
        assert isinstance(group, six.text_type), "%s is not text" % group
        assert isinstance(channel, six.text_type), "%s is not text" % channel
        # Remove from group set
        if group in self._groups:
            self._groups[group].discard(channel)
            if not self._groups[group]:
                del self._groups[group]

    def send_group(self, group, message):
        # Check types
        assert isinstance(message, dict), "message is not a dict"
        assert isinstance(group, six.text_type), "%s is not text" % group
        # Send to each channel
        for channel in self._groups.get(group, set()):
            self.send(channel, message)

    ### ASGI Flush API ###

    def flush(self):
        self._channels = {}
        self._groups = {}

    ### Expire cleanup ###

    def _clean_expired(self):
        """
        Goes through all messages and removes those that are expired.
        Any channel with an expired message is removed from all groups.
        """
        with self.thread_lock:
            for channel, queue in list(self._channels.items()):
                remove = False
                # See if it's expired
                while queue and queue[0][0] < time.time():
                    queue.popleft()
                    remove = True
                # Any removal prompts group discard
                if remove:
                    self._remove_from_groups(channel)
                # Is the channel now empty and needs deleting?
                if not queue:
                    del self._channels[channel]

    def _remove_from_groups(self, channel):
        """
        Removes a channel from all groups. Used when a message on it expires.
        """
        for  channels in self._groups.values():
            channels.discard(channel)

# Global single instance for easy use
channel_layer = ChannelLayer()
